const debugUrl = ''

// copied from Cocos engine souce code.
function visitAsset (asset: cc.Asset, deps: Object) {
    // assets generated programmatically or by user (e.g. label texture)  动态合图产生的纹理、字体纹理
    if (!asset['_uuid']) {
        if (asset instanceof cc.SpriteFrame) {
            const SpriteFrame = asset
            if (SpriteFrame['_original']) {
                const texture = SpriteFrame['_original']._texture;
                texture && deps[texture._uuid] = texture;
            } else {
                const texture = SpriteFrame.getTexture();
                texture && deps[texture._uuid] = texture;
            }
        }
        return;
    }
    deps[asset['_uuid']] = asset;
}

function visitComponent (comp: cc.Component, deps: Object)  {
    const props = Object.getOwnPropertyNames(comp);
    for (let i = 0; i < props.length; i++) {
        const propName = props[i];
        if (propName === 'node' || propName === '__eventTargets') { continue; }
        const value = comp[propName];
        if (value instanceof cc.NodePool) {
            visitNodePool(value, deps);
        } else 
        if (typeof value === 'object' && value) {
            if (Array.isArray(value)) {
                for (let j = 0; j < value.length; j++) {
                    const val = value[j];
                    if (val instanceof cc.Asset) {
                        visitAsset(val, deps);
                    }
                }
            } else if (!value.constructor || value.constructor === Object) {
                const keys = Object.getOwnPropertyNames(value);
                for (let j = 0; j < keys.length; j++) {
                    const val = value[keys[j]];
                    if (val instanceof cc.Asset) {
                        visitAsset(val, deps);
                    }
                }
            } else if (value instanceof cc.Asset) {
                visitAsset(value, deps);
            }
        }
    }
}

function visitNode (node: any, deps: Object) {
    for (let i = 0; i < node._components.length; i++) {
        visitComponent(node._components[i], deps);
    }
    for (let i = 0; i < node._children.length; i++) {
        visitNode(node._children[i], deps);
    }
}

function visitNodePool(nodePool: cc.NodePool, deps: Object) {
    const pool: Node[] = nodePool['_pool']
    for (let j = 0; j < pool.length; j++) {
        const node = pool[j];
        visitNode(node, deps);
    }
}

export class ResCleaner {
    /**
    * 资源清理 
    */
    static clean() {
        // return;
        cc.warn('--->资源释放 开始')
        let start = Date.now()

        let deps = {}
        let assetsCache = cc.assetManager.assets

        // 统计场景引用到的资源
        let nodeList = cc.director.getScene().children
        for (let i = 0; i < nodeList.length; i++) {
            visitNode(nodeList[i], deps)
        }

        // 不需要处理加载中的资源，引擎在加载中会自动加引用计数
        // 也不需要处理内置资源

        // 场景引用的资源引用加一
        for (const uuid in deps) {
            const asset = deps[uuid];
            asset.addRef()
        }

        let InitCount = assetsCache.count;
        cc.assetManager['releaseUnusedAssets']()

       setTimeout(() => {
            // 恢复资源的引用计数
            for (const uuid in deps) {
                const asset = deps[uuid];
                asset.decRef(false)
            }
            
           cc.warn('资源大约释放数量：', InitCount - assetsCache.count);
            // 尝试触发GC
            if (cc.sys.platform === cc.sys.WECHAT_GAME) {
                window['wx'].triggerGC && window['wx'].triggerGC();
            }
       }, 100);
    }
}

// 方便在控制台调用
window['cc'].ResCleaner = ResCleaner